Option Compare Binary
Option Explicit On
Option Strict On

Imports Microsoft.VisualBasic
Imports MvpsOrg.Dotnet.Win32.ResLib.NativeMethods
Imports System
Imports System.Collections.ObjectModel
Imports System.ComponentModel
Imports System.Drawing
Imports System.Globalization
Imports System.Runtime.InteropServices
Imports System.Text
Imports System.Windows.Forms

''' <summary>
'''   Provides information about the Win32 resources of an image file.
''' </summary>
Public NotInheritable Class ResourceManager
    Implements IDisposable

    Private m_FileName As String
    Private m_ModuleHandle As IntPtr

    ' Filled in callback functions.
    Private m_ResourceTypes As Collection(Of ResourceType)
    Private m_ResourceIds As Collection(Of ResourceId)
    Private m_ResourceLanguages As Collection(Of Short)

    Private m_Resources As Collection(Of ResourceInfo)

    ''' <summary>
    '''   Flags specifying which of the parameters of methods used to find
    '''   resources contain valid values.
    ''' </summary>
    <Flags()> _
    Public Enum FindResourceParameters

        ''' <summary>
        '''   The <paramref name="Id"/> parameter contains a valid value.
        ''' </summary>
        Id = 1

        ''' <summary>
        '''   The <paramref name="Type"/> parameter contains a valid value.
        ''' </summary>
        Type = 1 << 1

        ''' <summary>
        '''   The <paramref name="Language"/> parameter contains a valid
        '''   value.
        ''' </summary>
        Language = 1 << 2
    End Enum

    ''' <summary>Initializes a new instance of the class.</summary>
    Public Sub New()
        Me.ModuleHandle = IntPtr.Zero
    End Sub

    ''' <summary>Initializes a new instance of the class.</summary>
    ''' <param name="FileName">The file name of the image file.</param>
    ''' <param name="LoadAsDataFile">
    '''   Specifies whether the image file should be loaded as data file.
    '''   Loading the image file as data file will only allow low-level
    '''   resource access functions, loading bitmaps, icons, cursors,
    '''   strings, and messages will not work. The default value is
    '''   <c>False</c>.
    ''' </param>
    Public Sub New( _
        ByVal FileName As String, _
        Optional ByVal LoadAsDataFile As Boolean = False _
    )
        Me.FileName = FileName
        If LoadAsDataFile Then
            Me.ModuleHandle = _
                LoadLibraryEx( _
                    Me.FileName, _
                    IntPtr.Zero, _
                    LoadLibraryEx_Flags.LOAD_LIBRARY_AS_DATAFILE _
                )
        Else
            Me.ModuleHandle = LoadLibrary(Me.FileName)
        End If
        If Me.ModuleHandle = IntPtr.Zero Then
            Throw New Win32Exception()
        End If
    End Sub

    ''' <summary>Initializes a new instance of the class.</summary>
    ''' <param name="ModuleHandle">
    '''   The module handle identifying the image file.
    ''' </param>
    Public Sub New(ByVal ModuleHandle As IntPtr)
        Me.ModuleHandle = ModuleHandle
    End Sub

    ''' <summary>Gets or sets the file name of the image file.</summary>
    ''' <value>The file name of the image file.</value>
    ''' <returns>
    '''   A string containing the file name of the image file.
    ''' </returns>
    Public Property FileName() As String
        Get
            If m_FileName Is Nothing Then
                Me.FileName = _
                    GetModuleFileName(Me.ModuleHandle)
            End If
            Return m_FileName
        End Get
        Private Set(ByVal Value As String)
            m_FileName = Value
        End Set
    End Property

    ''' <summary>
    '''   Gets or sets the module handle identifying the image file.
    ''' </summary>
    ''' <value>The module handle identifying the image file.</value>
    ''' <returns>
    '''   A <c>System.IntPtr</c> containing the module handle identifying
    '''   the image file.
    ''' </returns>
    Public Property ModuleHandle() As IntPtr
        Get
            Return m_ModuleHandle
        End Get
        Private Set(ByVal Value As IntPtr)
            m_ModuleHandle = Value
        End Set
    End Property

    ''' <summary>
    '''   Loads a bitmap resource from the image file.
    ''' </summary>
    ''' <param name="Id">
    '''   The resource identifier of the bitmap resource to be loaded.
    ''' </param>
    ''' <returns>
    '''   A bitmap representing the value of the bitmap resource.
    ''' </returns>
    Public Function LoadBitmap(ByVal Id As ResourceId) As Bitmap
        Return _
            Bitmap.FromHbitmap( _
                LoadImage( _
                    Id, _
                    IMAGE_Constants.IMAGE_BITMAP, _
                    0, _
                    0, _
                    0 _
                ) _
            )
    End Function

    ''' <summary>
    '''   Loads a bitmap resource from the image file.
    ''' </summary>
    ''' <param name="Id">
    '''   The resource identifier of the bitmap resource to be loaded.
    ''' </param>
    ''' <param name="Width">
    '''   The desired width of the bitmap in pixels.
    ''' </param>
    ''' <param name="Height">
    '''   The desired height of the bitmap in pixels.
    ''' </param>
    ''' <param name="Options">Options used for loading the bitmap.</param>
    ''' <returns>
    '''   A bitmap representing the value of the bitmap resource.
    ''' </returns>
    Public Function LoadBitmap( _
        ByVal Id As ResourceId, _
        ByVal Width As Integer, _
        ByVal Height As Integer, _
        ByVal Options As LoadImageOptions _
    ) As Bitmap
        Return _
            Bitmap.FromHbitmap( _
                LoadImage( _
                    Id, _
                    IMAGE_Constants.IMAGE_BITMAP, _
                    Height, _
                    Width, _
                    Options _
                ) _
            )
    End Function

    ''' <summary>
    '''   Loads an icon resource from the image file.
    ''' </summary>
    ''' <param name="Id">
    '''   The resource identifier of the icon resource to be loaded.
    ''' </param>
    ''' <returns>
    '''   An icon representing the value of the icon resource.
    ''' </returns>
    Public Function LoadIcon(ByVal Id As ResourceId) As Icon
        Return _
            Icon.FromHandle( _
                LoadImage( _
                    Id, _
                    IMAGE_Constants.IMAGE_ICON, _
                    0, _
                    0, _
                    0 _
                ) _
            )
    End Function

    ''' <summary>
    '''    Loads an icon resource from the image file.
    ''' </summary>
    ''' <param name="Id">
    '''   The resource identifier of the icon resource to be loaded.
    ''' </param>
    ''' <param name="Width">
    '''   The desired width of the icon in pixels.
    ''' </param>
    ''' <param name="Height">
    '''   The desired height of the icon in pixels.
    ''' </param>
    ''' <param name="Options">Options used for loading the icon.</param>
    ''' <returns>
    '''   An icon representing the value of the icon resource.
    ''' </returns>
    Public Function LoadIcon( _
        ByVal Id As ResourceId, _
        ByVal Width As Integer, _
        ByVal Height As Integer, _
        ByVal Options As LoadImageOptions _
    ) As Icon
        Return _
            Icon.FromHandle( _
                LoadImage( _
                    Id, _
                    IMAGE_Constants.IMAGE_ICON, _
                    Height, _
                    Width, _
                    Options _
                ) _
            )
    End Function

    ''' <summary>
    '''    Loads a cursor resource from the image file.
    ''' </summary>
    ''' <param name="Id">
    '''   The resource identifier of the cursor resource to be loaded.
    ''' </param>
    ''' <returns>
    '''   A cursor representing the value of the cursor resource.
    ''' </returns>
    Public Function LoadCursor(ByVal Id As ResourceId) As Cursor
        Return _
            New Cursor( _
                LoadImage( _
                    Id, _
                    IMAGE_Constants.IMAGE_CURSOR, _
                    0, _
                    0, _
                    0 _
                ) _
            )
    End Function

    ''' <summary>
    '''   Loads a cursor resource from the image file.
    ''' </summary>
    ''' <param name="Id">
    '''   The resource identifier of the cursor resource to be loaded.
    ''' </param>
    ''' <param name="Width">
    '''   The desired width of the cursor in pixels.
    ''' </param>
    ''' <param name="Height">
    '''   The desired height of the cursor in pixels.
    ''' </param>
    ''' <param name="Options">Options used for loading the cursor.</param>
    ''' <returns>
    '''   A cursor representing the value of the cursor resource.
    ''' </returns>
    Public Function LoadCursor( _
        ByVal Id As ResourceId, _
        ByVal Width As Integer, _
        ByVal Height As Integer, _
        ByVal Options As LoadImageOptions _
    ) As Cursor
        Return _
            New Cursor( _
                LoadImage( _
                    Id, _
                    IMAGE_Constants.IMAGE_CURSOR, _
                    Height, _
                    Width, _
                    Options _
                ) _
            )
    End Function

    ''' <summary>
    '''   Loads a string resource from the image file.
    ''' </summary>
    ''' <param name="Id">
    '''   The integer identifier of the string resource.
    ''' </param>
    ''' <returns>
    '''   A string containing the value of the string resource.
    ''' </returns>
    Public Function LoadString(ByVal Id As Integer) As String

        ' &H1000 is the maximum length of a string-table entry.
        Dim Buffer As String = Space(&H1000)
        Dim n As Int32 = _
            NativeMethods.LoadString( _
                Me.ModuleHandle, _
                Id, _
                Buffer, _
                Buffer.Length _
            )
        If n = 0 Then
            Throw New Win32Exception()
        End If
        Return Left(Buffer, n)
    End Function

    ''' <summary>
    '''   Loads a message resource from the image file.
    ''' </summary>
    ''' <param name="Id">
    '''   The integer identifier of the message resource.
    ''' </param>
    ''' <param name="IgnoreRegularLineBreaks">
    '''   Specifies whether regular line breaks should not be put into the
    '''   output. Hard-coded line breaks (encoded by <c>%n</c>) are copied
    '''   to the output. The default value is <c>False</c>.
    ''' </param>
    ''' <returns>
    '''   A string containing the value of the message resource.
    ''' </returns>
    Public Function LoadMessage( _
        ByVal Id As Integer, _
        Optional ByVal IgnoreRegularLineBreaks As Boolean = False _
    ) As String
        Return Me.LoadMessage(Id, NeutralLanguage, IgnoreRegularLineBreaks)
    End Function

    ''' <summary>
    '''   Loads a message resource from the image file.
    ''' </summary>
    ''' <param name="Id">
    '''   The integer identifier of the message resource.
    ''' </param>
    ''' <param name="MaximumLineWidth">
    '''   The maximum number of characters in an output line of the
    '''   formatted message.
    ''' </param>
    ''' <returns>
    '''   A string containing the value of the message resource.
    ''' </returns>
    Public Function LoadMessage( _
        ByVal Id As Integer, _
        ByVal MaximumLineWidth As Integer _
    ) As String
        Return Me.LoadMessage(Id, NeutralLanguage, MaximumLineWidth)
    End Function

    ''' <summary>
    '''   Loads a message resource from the image file.
    ''' </summary>
    ''' <param name="Id">
    '''   The integer identifier of the message resource.
    ''' </param>
    ''' <param name="Language">
    '''   The language of the message resource.
    ''' </param>
    ''' <param name="IgnoreRegularLineBreaks">
    '''   Specifies whether regular line breaks should not be put into the
    '''   output. Hard-coded line breaks (encoded by <c>%n</c>) are copied
    '''   to the output. The default value is <c>False</c>.
    ''' </param>
    ''' <returns>
    '''   A string containing the value of the message resource.
    ''' </returns>
    Public Function LoadMessage( _
        ByVal Id As Integer, _
        ByVal Language As Short, _
        Optional ByVal IgnoreRegularLineBreaks As Boolean = False _
    ) As String
        Dim Flags As FORMAT_MESSAGE_Flags = _
            FORMAT_MESSAGE_Flags.FORMAT_MESSAGE_FROM_HMODULE Or _
            FORMAT_MESSAGE_Flags.FORMAT_MESSAGE_ALLOCATE_BUFFER Or _
            FORMAT_MESSAGE_Flags.FORMAT_MESSAGE_IGNORE_INSERTS
        If IgnoreRegularLineBreaks Then
            Flags = Flags Or FORMAT_MESSAGE_Flags.FORMAT_MESSAGE_MAX_WIDTH_MASK
        End If
        Return Me.LoadMessage(Id, Language, Flags, Nothing)
    End Function

    ''' <summary>
    '''   Loads a message resource from the image file.
    ''' </summary>
    ''' <param name="Id">
    '''   The integer identifier of the message resource.
    ''' </param>
    ''' <param name="Language">
    '''   The language of the message resource.
    ''' </param>
    ''' <param name="MaximumLineWidth">
    '''   The maximum number of characters in an output line of the
    '''   formatted message.
    ''' </param>
    ''' <returns>
    '''   A string containing the value of the message resource.
    ''' </returns>
    Public Function LoadMessage( _
        ByVal Id As Integer, _
        ByVal Language As Short, _
        ByVal MaximumLineWidth As Integer _
    ) As String

        ' 255 means 'FORMAT_MESSAGE_MAX_WIDTH_MASK' and is thus not
        ' allowed.  0 means that no line breaks are inserted and thus is
        ' not allowed.
        If MaximumLineWidth < 1 OrElse MaximumLineWidth > 254 Then
            Throw New ArgumentException(String.Format(CultureInfo.InvariantCulture, My.Resources.ExceptionParameterXMustBeInRangeYToZ, "MaximumLineWidth", 1, 254))
        End If

        Dim Flags As FORMAT_MESSAGE_Flags = _
            FORMAT_MESSAGE_Flags.FORMAT_MESSAGE_FROM_HMODULE Or _
            FORMAT_MESSAGE_Flags.FORMAT_MESSAGE_ALLOCATE_BUFFER Or _
            FORMAT_MESSAGE_Flags.FORMAT_MESSAGE_ARGUMENT_ARRAY Or _
            CType(MaximumLineWidth, FORMAT_MESSAGE_Flags)
        Return Me.LoadMessage(Id, Language, Flags, Nothing)
    End Function

    ''' <summary>
    '''   Loads a message resource from the image file.
    ''' </summary>
    ''' <param name="Id">
    '''   The integer identifier of the message resource.
    ''' </param>
    ''' <param name="Arguments">
    '''   The arguments to be inserted into the message.
    ''' </param>
    ''' <param name="IgnoreRegularLineBreaks">
    '''   Specifies whether regular line breaks should not be put into the
    '''   output. Hard-coded line breaks (encoded by <c>%n</c>) are copied
    '''   to the output. The default value is <c>False</c>.
    ''' </param>
    ''' <returns>
    '''   A string containing the value of the message resource.
    ''' </returns>
    Public Function LoadMessage( _
        ByVal Id As Integer, _
        ByVal Arguments() As String, _
        Optional ByVal IgnoreRegularLineBreaks As Boolean = False _
    ) As String
        Return _
            Me.LoadMessage( _
                Id, _
                NeutralLanguage, _
                Arguments, _
                IgnoreRegularLineBreaks _
            )
    End Function

    ''' <summary>
    '''   Loads a message resource from the image file.
    ''' </summary>
    ''' <param name="Id">
    '''   The integer identifier of the message resource.
    ''' </param>
    ''' <param name="Arguments">
    '''   The arguments to be inserted into the message.
    ''' </param>
    ''' <param name="MaximumLineWidth">
    '''   The maximum number of characters in an output line of the
    '''   formatted message.
    ''' </param>
    ''' <returns>
    '''   A string containing the value of the message resource.
    ''' </returns>
    Public Function LoadMessage( _
        ByVal Id As Integer, _
        ByVal Arguments() As String, _
        ByVal MaximumLineWidth As Integer _
    ) As String
        Return _
            Me.LoadMessage( _
                Id, _
                NeutralLanguage, _
                Arguments, _
                MaximumLineWidth _
            )
    End Function

    ''' <summary>
    '''   Loads a message resource from the image file.
    ''' </summary>
    ''' <param name="Id">
    '''   The integer identifier of the message resource.
    ''' </param>
    ''' <param name="Language">
    '''   The language of the message resource.
    ''' </param>
    ''' <param name="Arguments">
    '''   The arguments to be inserted into the message.
    ''' </param>
    ''' <param name="IgnoreRegularLineBreaks">
    '''   Specifies whether regular line breaks should not be put into the
    '''   output. Hard-coded line breaks (encoded by <c>%n</c>) are copied
    '''   to the output. The default value is <c>False</c>.
    ''' </param>
    ''' <returns>
    '''   A string containing the value of the message resource.
    ''' </returns>
    Public Function LoadMessage( _
        ByVal Id As Integer, _
        ByVal Language As Short, _
        ByVal Arguments() As String, _
        Optional ByVal IgnoreRegularLineBreaks As Boolean = False _
    ) As String
        If Arguments Is Nothing Then
            Throw New ArgumentException(String.Format(CultureInfo.InvariantCulture, My.Resources.ExceptionParameterXShouldHaveValidValueButIsNothing, "Arguments"))
        End If
        Dim Flags As FORMAT_MESSAGE_Flags = _
            FORMAT_MESSAGE_Flags.FORMAT_MESSAGE_FROM_HMODULE Or _
            FORMAT_MESSAGE_Flags.FORMAT_MESSAGE_ALLOCATE_BUFFER Or _
            FORMAT_MESSAGE_Flags.FORMAT_MESSAGE_ARGUMENT_ARRAY
        If IgnoreRegularLineBreaks Then
            Flags = Flags Or FORMAT_MESSAGE_Flags.FORMAT_MESSAGE_MAX_WIDTH_MASK
        End If
        Return Me.LoadMessage(Id, Language, Flags, Arguments)
    End Function

    ''' <summary>
    '''   Loads a message resource from the image file.
    ''' </summary>
    ''' <param name="Id">
    '''   The integer identifier of the message resource.
    ''' </param>
    ''' <param name="Language">
    '''   The language of the message resource.
    ''' </param>
    ''' <param name="Arguments">
    '''   The arguments to be inserted into the message.
    ''' </param>
    ''' <param name="MaximumLineWidth">
    '''   The maximum number of characters in an output line of the
    '''   formatted message.
    ''' </param>
    ''' <returns>
    '''   A string containing the value of the message resource.
    ''' </returns>
    Public Function LoadMessage( _
        ByVal Id As Integer, _
        ByVal Language As Short, _
        ByVal Arguments() As String, _
        ByVal MaximumLineWidth As Integer _
    ) As String
        If Arguments Is Nothing Then
            Throw New ArgumentException(String.Format(CultureInfo.InvariantCulture, My.Resources.ExceptionParameterXShouldHaveValidValueButIsNothing, "Arguments"))
        End If

        ' 255 means 'FORMAT_MESSAGE_MAX_WIDTH_MASK' and is thus not
        ' allowed.  0 means that no line breaks are inserted and thus is
        ' not allowed.
        If MaximumLineWidth < 1 OrElse MaximumLineWidth > 254 Then
            Throw New ArgumentException(String.Format(CultureInfo.InvariantCulture, My.Resources.ExceptionParameterXMustBeInRangeYToZ, "MaximumLineWidth", 1, 254))
        End If

        Dim Flags As FORMAT_MESSAGE_Flags = _
            FORMAT_MESSAGE_Flags.FORMAT_MESSAGE_FROM_HMODULE Or _
            FORMAT_MESSAGE_Flags.FORMAT_MESSAGE_ALLOCATE_BUFFER Or _
            FORMAT_MESSAGE_Flags.FORMAT_MESSAGE_ARGUMENT_ARRAY Or _
            CType(MaximumLineWidth, FORMAT_MESSAGE_Flags)
        Return Me.LoadMessage(Id, Language, Flags, Arguments)
    End Function

    ''' <summary>
    '''   Formats a message by inserting arguments according to the format
    '''   specified in the message.
    ''' </summary>
    ''' <param name="Message">The message to format.</param>
    ''' <param name="Arguments">
    '''   The arguments to be inserted into the message.
    ''' </param>
    ''' <param name="IgnoreRegularLineBreaks">
    '''   Specifies whether regular line breaks should not be put into the
    '''   output. Hard-coded line breaks (encoded by <c>%n</c>) are copied
    '''   to the output. The default value is <c>False</c>.
    ''' </param>
    ''' <returns>A string containing the formatted message.</returns>
    Public Shared Function FormatMessage( _
        ByVal Message As String, _
        ByVal Arguments() As String, _
        Optional ByVal IgnoreRegularLineBreaks As Boolean = False _
    ) As String
        If Arguments Is Nothing Then
            Throw New ArgumentException(String.Format(CultureInfo.InvariantCulture, My.Resources.ExceptionParameterXShouldHaveValidValueButIsNothing, "Arguments"))
        End If
        Dim Flags As FORMAT_MESSAGE_Flags = _
            FORMAT_MESSAGE_Flags.FORMAT_MESSAGE_ALLOCATE_BUFFER Or _
            FORMAT_MESSAGE_Flags.FORMAT_MESSAGE_FROM_STRING Or _
            FORMAT_MESSAGE_Flags.FORMAT_MESSAGE_ARGUMENT_ARRAY
        If IgnoreRegularLineBreaks Then
            Flags = Flags Or FORMAT_MESSAGE_Flags.FORMAT_MESSAGE_MAX_WIDTH_MASK
        End If
        Return FormatMessage(Message, Arguments, Flags)
    End Function

    ''' <summary>
    '''   Formats a message by inserting arguments according to the format
    '''   specified in the message.
    ''' </summary>
    ''' <param name="Message">The message to format.</param>
    ''' <param name="Arguments">
    '''   The arguments to be inserted into the message.
    ''' </param>
    ''' <param name="MaximumLineWidth">
    '''   The maximum number of characters in an output line of the
    '''   formatted message.
    ''' </param>
    ''' <returns>A string containing the formatted message.</returns>
    Public Shared Function FormatMessage( _
        ByVal Message As String, _
        ByVal Arguments() As String, _
        ByVal MaximumLineWidth As Integer _
    ) As String
        If Arguments Is Nothing Then
            Throw New ArgumentException(String.Format(CultureInfo.InvariantCulture, My.Resources.ExceptionParameterXShouldHaveValidValueButIsNothing, "Arguments"))
        End If

        ' 255 means 'FORMAT_MESSAGE_MAX_WIDTH_MASK' and is thus not
        ' allowed.  0 means that no line breaks are inserted and thus is
        ' not allowed.
        If MaximumLineWidth < 1 OrElse MaximumLineWidth > 254 Then
            Throw New ArgumentException(String.Format(CultureInfo.InvariantCulture, My.Resources.ExceptionParameterXMustBeInRangeYToZ, "MaximumLineWidth", 1, 254))
        End If

        Dim Flags As FORMAT_MESSAGE_Flags = _
            FORMAT_MESSAGE_Flags.FORMAT_MESSAGE_ALLOCATE_BUFFER Or _
            FORMAT_MESSAGE_Flags.FORMAT_MESSAGE_FROM_STRING Or _
            FORMAT_MESSAGE_Flags.FORMAT_MESSAGE_ARGUMENT_ARRAY Or _
            CType(MaximumLineWidth, FORMAT_MESSAGE_Flags)
        Return FormatMessage(Message, Arguments, Flags)
    End Function

    ''' <summary>
    '''   Determines the resource types in the image file and returns them
    '''   in a list.
    ''' </summary>
    ''' <returns>
    '''   A list of resource types containing the resource types found in
    '''   the image file.
    ''' </returns>
    Public Function GetResourceTypes() As Collection(Of ResourceType)
        m_ResourceTypes = New Collection(Of ResourceType)()
        If _
            Not EnumResourceTypes( _
                Me.ModuleHandle, _
                AddressOf EnumResTypeProc, _
                IntPtr.Zero _
            ) _
        Then
            Throw New Win32Exception()
        End If
        Return m_ResourceTypes
    End Function

    Private Function EnumResTypeProc( _
        ByVal hModule As IntPtr, _
        ByVal lpszType As IntPtr, _
        ByVal lParam As IntPtr _
    ) As Boolean
        m_ResourceTypes.Add(ResourceType.FromLPStr(lpszType))

        ' Continue enumeration.
        Return True
    End Function

    ''' <summary>
    '''   Determines the resource identifiers in the image file of the
    '''   specified resource type and returns them in a list.
    ''' </summary>
    ''' <param name="Type">
    '''   The type of the resources whose identifiers should be determined.
    ''' </param>
    ''' <returns>
    '''   A list of resource identifiers containing the resource
    '''   identifiers of the specified type found in the image file.
    ''' </returns>
    Public Function GetResourceIds( _
        ByVal Type As ResourceType _
    ) As Collection(Of ResourceId)
        m_ResourceIds = New Collection(Of ResourceId)()
        If _
            Not EnumResourceNames( _
                Me.ModuleHandle, _
                Type, _
                AddressOf EnumResNameProc, _
                IntPtr.Zero _
            ) _
        Then
            Dim Err As Integer = Marshal.GetLastWin32Error()
            If _
                Err <> NO_ERROR AndAlso _
                Err <> ERROR_RESOURCE_TYPE_NOT_FOUND _
            Then
                Throw New Win32Exception()
            End If
        End If
        Return m_ResourceIds
    End Function

    Private Function EnumResNameProc( _
        ByVal hModule As IntPtr, _
        ByVal lpszType As IntPtr, _
        ByVal lpszName As IntPtr, _
        ByVal lParam As IntPtr _
    ) As Boolean
        m_ResourceIds.Add(ResourceId.FromLPStr(lpszName))

        ' Continue enumeration.
        Return True
    End Function

    ''' <summary>
    '''   Determines the resource languages in the image file of the
    '''   specified resource type and resource identifier and returns them
    '''   in a list.
    ''' </summary>
    ''' <param name="Type">
    '''   The resource type of the resources whose languages should be
    '''   determined.
    ''' </param>
    ''' <param name="Id">
    '''   The resource identifier of the resources whose languages should
    '''   be determined.
    ''' </param>
    ''' <returns>
    '''   A list of languages containing the languages of the resources of
    '''   the specified type and identifier found in the image file.
    ''' </returns>
    Public Function GetResourceLanguages( _
        ByVal Type As ResourceType, _
        ByVal Id As ResourceId _
    ) As Collection(Of Short)
        m_ResourceLanguages = New Collection(Of Short)()
        If _
            Not EnumResourceLanguages( _
                Me.ModuleHandle, _
                Type, _
                Id, _
                AddressOf EnumResLangProc, _
                IntPtr.Zero _
            ) _
        Then
            Throw New Win32Exception()
        End If
        Return m_ResourceLanguages
    End Function

    Private Function EnumResLangProc( _
        ByVal hModule As IntPtr, _
        ByVal lpszType As IntPtr, _
        ByVal lpszName As IntPtr, _
        ByVal wIDLanguage As Int16, _
        ByVal lParam As IntPtr _
    ) As Boolean
        m_ResourceLanguages.Add(wIDLanguage)

        ' Continue enumeration.
        Return True
    End Function

    ''' <summary>
    '''   Determines whether or not the image file contains a resource
    '''   identified by the specified values.
    ''' </summary>
    ''' <param name="Type">The resource type of the resource.</param>
    ''' <param name="Id">The resource identifier of the resource.</param>
    ''' <returns>
    '''   <c>True</c> if the resource identified by the specified values
    '''   could be found in the image file; otherwise, <c>False</c>.
    ''' </returns>
    Public Function ContainsResource( _
        ByVal Type As ResourceType, _
        ByVal Id As ResourceId _
    ) As Boolean
        Dim hResource As IntPtr = _
            FindResource(Me.ModuleHandle, Id, Type)
        Return hResource <> IntPtr.Zero
    End Function

    ''' <summary>
    '''   Determines whether or not the image file contains a resource
    '''   identified by the specified values.
    ''' </summary>
    ''' <param name="Type">The resource type of the resource.</param>
    ''' <param name="Id">The resource identifier of the resource.</param>
    ''' <param name="Language">
    '''   The resource language of the resource.
    ''' </param>
    ''' <returns>
    '''   <c>True</c> if the resource identified by the specified values
    '''   could be found in the image file; otherwise, <c>False</c>.
    ''' </returns>
    Public Function ContainsResource( _
        ByVal Type As ResourceType, _
        ByVal Id As ResourceId, _
        ByVal Language As Short _
    ) As Boolean
        Dim hResource As IntPtr = _
            FindResourceEx(Me.ModuleHandle, Type, Id, Language)
        Return hResource <> IntPtr.Zero
    End Function

    ''' <summary>
    '''   Searches the image file for resources meeting the specified
    '''   criteria and returns them in a list.
    ''' </summary>
    ''' <param name="FindAttrs">
    '''   Attributes specifying which of the other parameters contain valid
    '''   values.
    ''' </param>
    ''' <param name="Type">The type of the resources to be found.</param>
    ''' <param name="Id">
    '''   The identifier of the resources to be found.
    ''' </param>
    ''' <param name="Language">
    '''   The language of the resources to be found.
    ''' </param>
    ''' <returns>
    '''   A list of resource information objects containing the resource
    '''   information objects meeting the criteria.
    ''' </returns>
    Public Function FindResources( _
        ByVal FindAttrs As FindResourceParameters, _
        ByVal Id As ResourceId, _
        ByVal Type As ResourceType, _
        ByVal Language As Short _
    ) As Collection(Of ResourceInfo)
        Me.MakeSureResourcesAreInitialized()
        Dim ById As Boolean = CBool(FindAttrs And FindResourceParameters.Id)
        If ById AndAlso Id Is Nothing Then
            Throw New ArgumentException(String.Format(CultureInfo.InvariantCulture, My.Resources.ExceptionParameterXShouldHaveValidValueButIsNothing, "Id"))
        End If
        Dim ByType As Boolean = CBool(FindAttrs And FindResourceParameters.Type)
        If ByType AndAlso Type Is Nothing Then
            Throw New ArgumentException(String.Format(CultureInfo.InvariantCulture, My.Resources.ExceptionParameterXShouldHaveValidValueButIsNothing, "Type"))
        End If
        Dim ByLanguage As Boolean = _
            CBool(FindAttrs And FindResourceParameters.Language)
        Dim Resources As New Collection(Of ResourceInfo)()
        For Each Resource As ResourceInfo In m_Resources
            Dim Interesting As Boolean = True
            If ById AndAlso Resource.Id <> Id Then
                Interesting = False
            End If
            If ByType AndAlso Resource.Type <> Type Then
                Interesting = False
            End If
            If ByLanguage AndAlso Resource.Language <> Language Then
                Interesting = False
            End If
            If Interesting Then
                Resources.Add(Resource)
            End If
        Next Resource
        Return Resources
    End Function

    ''' <summary>
    '''   Searches the image file for the first resource which meets the
    '''   specified criteria and returns this resource.
    ''' </summary>
    ''' <param name="Type">The type of the resource to be found.</param>
    ''' <param name="Id">
    '''   The identifier of the resource to be found.
    ''' </param>
    ''' <param name="Language">
    '''   The language of the resource to be found.
    ''' </param>
    ''' <returns>
    '''   A resource information object if a resource meeting the specified
    '''   criteria could be found; otherwise <c>Nothing</c>.
    ''' </returns>
    Public Function FindSingleResource( _
         ByVal Type As ResourceType, _
         ByVal Id As ResourceId, _
         ByVal Language As Short _
    ) As ResourceInfo
        Me.MakeSureResourcesAreInitialized()
        For Each Resource As ResourceInfo In m_Resources
            If _
                Resource.Type = Type AndAlso _
                Resource.Id = Id AndAlso _
                Resource.Language = Language _
            Then
                Return Resource
            End If
        Next Resource
        Return Nothing
    End Function

    ''' <summary>
    '''   Searches the image file for the first resource which meets the
    '''   specified criteria and returns this resource.
    ''' </summary>
    ''' <param name="Type">The type of the resource to be found.</param>
    ''' <param name="Id">
    '''   The identifier of the resource to be found.
    ''' </param>
    ''' <returns>
    '''   A resource information object if a resource meeting the specified
    '''   criteria could be found; otherwise <c>Nothing</c>.
    ''' </returns>
    Public Function FindSingleResource( _
        ByVal Type As ResourceType, _
        ByVal Id As ResourceId _
    ) As ResourceInfo
        Me.MakeSureResourcesAreInitialized()
        For Each Resource As ResourceInfo In m_Resources
            If Resource.Type = Type AndAlso Resource.Id = Id Then
                Return Resource
            End If
        Next Resource
        Return Nothing
    End Function

    ''' <summary>Returns a string representing the object.</summary>
    ''' <returns>A string representing the object.</returns>
    Public Overrides Function ToString() As String
        Dim Value As New StringBuilder()
        For Each Type As ResourceType In Me.GetResourceTypes()
            Value.AppendLine("Type: " & Type.ToString())
            For Each Name As ResourceId In Me.GetResourceIds(Type)
                Value.AppendLine("  ID: " & Name.ToString())
                For Each Language As Short In Me.GetResourceLanguages(Type, Name)
                    Value.AppendLine("    Language: " & LangIdToTwoLetterCode(Language))
                Next Language
            Next Name
        Next Type
        Return Value.ToString()
    End Function

    ''' <summary>
    '''   Releases the unmanaged resources used by the instance.
    ''' </summary>
    Public Sub Dispose() Implements IDisposable.Dispose
        If Me.ModuleHandle <> IntPtr.Zero Then
            If Not FreeLibrary(Me.ModuleHandle) Then
                Throw New Win32Exception()
            End If
            Me.ModuleHandle = IntPtr.Zero
        End If
    End Sub

    Private Function GetResources() As Collection(Of ResourceInfo)
        Dim Resources As New Collection(Of ResourceInfo)()
        For Each Type As ResourceType In Me.GetResourceTypes()
            For Each Id As ResourceId In Me.GetResourceIds(Type)
                For Each Language As Short In Me.GetResourceLanguages(Type, Id)
                    Resources.Add(New ResourceInfo(Type, Id, Language, Me.ModuleHandle))
                Next Language
            Next Id
        Next Type
        Return Resources
    End Function

    Private Sub MakeSureResourcesAreInitialized()
        If m_Resources Is Nothing Then
            m_Resources = GetResources()
        End If
    End Sub

    Private Function LoadImage( _
        ByVal Id As ResourceId, _
        ByVal Type As IMAGE_Constants, _
        ByVal Width As Integer, _
        ByVal Height As Integer, _
        ByVal Options As LoadImageOptions _
    ) As IntPtr
        Dim h As IntPtr = _
            NativeMethods.LoadImage( _
                Me.ModuleHandle, _
                Id, _
                Type, _
                Width, _
                Height, _
                CType(Options, LR_Flags) _
            )
        If h = IntPtr.Zero Then
            Throw New Win32Exception()
        End If
        Return h
    End Function

    Private Function LoadMessage( _
        ByVal Id As Integer, _
        ByVal Language As Short, _
        ByVal Flags As FORMAT_MESSAGE_Flags, _
        ByVal Arguments() As String _
    ) As String
        Dim Buffer As IntPtr
        Dim n As Int32 = _
            NativeMethods.FormatMessage( _
                Flags, _
                Me.ModuleHandle, _
                Id, _
                Language, _
                Buffer, _
                0, _
                Arguments _
            )
        If n = 0 Then
            Throw New Win32Exception()
        End If
        Dim s As String = Marshal.PtrToStringAuto(Buffer, n)
        Marshal.FreeHGlobal(Buffer)
        Return s
    End Function

    Private Shared Function FormatMessage( _
        ByVal Message As String, _
        ByVal Arguments() As String, _
        ByVal Flags As FORMAT_MESSAGE_Flags _
    ) As String
        Dim Buffer As IntPtr
        Dim n As Int32 = _
            NativeMethods.FormatMessage( _
                Flags, _
                Message, _
                0, _
                0, _
                Buffer, _
                0, _
                Arguments _
            )
        If n = 0 Then
            Throw New Win32Exception()
        End If
        Dim s As String = Marshal.PtrToStringAuto(Buffer, n)
        Marshal.FreeHGlobal(Buffer)
        Return s
    End Function
End Class