Option Compare Binary
Option Explicit On 
Option Strict On

Imports System
Imports System.Collections.Generic
Imports System.ComponentModel
Imports System.Globalization
Imports System.Runtime.InteropServices

''' <summary>
'''   Provides functionality to update Win32 resources of an image file.
''' </summary>
Public NotInheritable Class ResourceUpdater
    Implements IDisposable

    Private m_hUpdateRes As IntPtr
    Private m_FileName As String

    ''' <summary>
    '''   Initializes a new instance of the class.
    ''' </summary>
    ''' <param name="FileName">
    '''   The name of the image file whose resources should be updated.
    ''' </param>
    Public Sub New(ByVal FileName As String)
        m_FileName = FileName
    End Sub

    ''' <summary>
    '''   Gets or sets the name of the image file whose resources are
    '''   updated.
    ''' </summary>
    ''' <value>
    '''   The file name of the image file whose resources are updated.
    ''' </value>
    ''' <returns>
    '''   A string containing the file name of the image file whose
    '''   resources are updated.
    ''' </returns>
    Public Property FileName() As String
        Get
            Return m_FileName
        End Get
        Private Set(ByVal Value As String)
            m_FileName = Value
        End Set
    End Property

    ''' <summary>
    '''   Opens the image file for updating the resources.
    ''' </summary>
    ''' <param name="DeleteExistingResources">
    '''   Determines whether existing resources in the image file should be
    '''   deleted. The default value is <c>False</c>.
    ''' </param>
    ''' <remarks>
    '''   Calls to <c>BeginUpdate</c> and <c>EndUpdate</c> must be paired.
    ''' </remarks>
    ''' <exception cref="InvalidOperationException">
    '''   The image file has already been opened for updating.
    ''' </exception>
    Public Sub BeginUpdate( _
        Optional ByVal DeleteExistingResources As Boolean = False _
    )
        If m_hUpdateRes <> IntPtr.Zero Then
            Throw New InvalidOperationException(My.Resources.ExceptionImageFileAlreadyOpenedForUpdating)
        End If
        m_hUpdateRes = _
            NativeMethods.BeginUpdateResource( _
                Me.FileName, _
                DeleteExistingResources _
            )
        If m_hUpdateRes = IntPtr.Zero Then
            Throw New Win32Exception()
        End If
    End Sub

    ''' <summary>
    '''   Finishes updating the resources by performing the updates in the
    '''   image file.
    ''' </summary>
    ''' <param name="DiscardChanges">
    '''   Determines whether the updates made to the resources should be
    '''   performed in the image file. The default value is <c>False</c>.
    ''' </param>
    ''' <remarks>
    '''   Calls to <c>BeginUpdate</c> and <c>EndUpdate</c> must be paired.
    ''' </remarks>
    ''' <exception cref="InvalidOperationException">
    '''   The image file has not been opened for updating yet.
    ''' </exception>
    Public Sub EndUpdate( _
        Optional ByVal DiscardChanges As Boolean = False _
    )
        If m_hUpdateRes = IntPtr.Zero Then
            Throw New InvalidOperationException(My.Resources.ExceptionImageFileNotOpenedForUpdating)
        End If
        If _
            Not NativeMethods.EndUpdateResource( _
                m_hUpdateRes, _
                DiscardChanges _
            ) _
        Then
            Throw New Win32Exception()
        End If
        m_hUpdateRes = IntPtr.Zero
    End Sub

    ''' <summary>
    '''   Updates a resource of the image file or adds a resource to the
    '''   image file.
    ''' </summary>
    ''' <param name="Resource">
    '''   The resource to get updated or added.
    ''' </param>
    ''' <exception cref="InvalidOperationException">
    '''   The image file has not been opened for updating yet.
    ''' </exception>
    Public Sub UpdateResource(ByVal Resource As ResourceInfo)
        If m_hUpdateRes = IntPtr.Zero Then
            Throw New InvalidOperationException(My.Resources.ExceptionImageFileNotOpenedForUpdating)
        End If
        If Resource Is Nothing Then
            Throw New ArgumentException(String.Format(CultureInfo.InvariantCulture, My.Resources.ExceptionParameterXMustNotBeNothing, "Resource"))
        End If
        Dim Data() As Byte = Resource.GetData()
        If _
            Not NativeMethods.UpdateResource( _
                m_hUpdateRes, _
                Resource.Type, _
                Resource.Id, _
                Resource.Language, _
                Data, _
                Data.Length _
            ) _
        Then
            Throw New Win32Exception()
        End If
    End Sub

    ''' <summary>
    '''   Removes a resource from the image file.
    ''' </summary>
    ''' <param name="Resource">
    '''   The resource to get removed from the image file.
    ''' </param>
    ''' <exception cref="InvalidOperationException">
    '''   The image file has not been opened for updating yet.
    ''' </exception>
    Public Sub DeleteResource(ByVal Resource As ResourceInfo)
        If m_hUpdateRes = IntPtr.Zero Then
            Throw New InvalidOperationException(My.Resources.ExceptionImageFileNotOpenedForUpdating)
        End If
        If Resource Is Nothing Then
            Throw New ArgumentException(String.Format(CultureInfo.InvariantCulture, My.Resources.ExceptionParameterXMustNotBeNothing, "Resource"))
        End If
        If _
            Not NativeMethods.UpdateResource( _
                m_hUpdateRes, _
                Resource.Type, _
                Resource.Id, _
                Resource.Language, _
                Nothing, _
                0 _
            ) _
        Then
            Throw New Win32Exception()
        End If
    End Sub

    ''' <summary>
    '''   Releases the unmanaged resources used by the instance.
    ''' </summary>
    Public Sub Dispose() Implements IDisposable.Dispose
        If m_hUpdateRes <> IntPtr.Zero Then
            Me.EndUpdate()
        End If
    End Sub
End Class